/**
* Copyright (C) 2007 EDIT
* European Distributed Institute of Taxonomy 
* http://www.e-taxonomy.eu
* 
* The contents of this file are subject to the Mozilla Public License Version 1.1
* See LICENSE.TXT at the top of this package for the full license terms.
*/

package eu.etaxonomy.taxeditor.editor;

import org.apache.log4j.Logger;
import org.eclipse.jface.text.DocumentEvent;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.IDocumentListener;
import org.eclipse.jface.text.TextViewer;
import org.eclipse.swt.custom.StyledText;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.FocusListener;
import org.eclipse.swt.graphics.Font;

import eu.etaxonomy.taxeditor.ITaxEditorConstants;
import eu.etaxonomy.taxeditor.TaxEditorPlugin;

/**
 * Displays a message such as "Click here to start entering
 * data" when <code>document</code> is empty.
 * 
 * @author p.ciardelli
 * @created 15.09.2008
 * @version 1.0
 */
public class EmptyTextViewerPrompt implements IDocumentListener {
	private static final Logger logger = Logger
			.getLogger(EmptyTextViewerPrompt.class);
	
	private IDocument document;
	private StyledText textControl;
	private String prompt;
	private FocusListener focusListener = null;

	private Font textFont;
	private Font promptFont = TaxEditorPlugin.getDefault()
			.getFont(ITaxEditorConstants.DEFAULT_PROMPT_FONT);
	
	public EmptyTextViewerPrompt(TextViewer textViewer, String prompt) {

		this.document = textViewer.getDocument();
		this.textControl = textViewer.getTextWidget();
		this.prompt = prompt;

		this.textFont = textControl.getFont();
		
		documentChanged(null);		
	}

	/**
	 * Override <code>prompt</code>'s default font. 
	 * 
	 * @param font
	 */
	public void setPromptFont(Font font) {
		this.promptFont = font;
	}
	
	/**
	 * Adds a focus listener to the <code>textControl</code>, which hides 
	 * the <code>prompt</code> if the cursor is in the <code>textControl</code>, 
	 * and shows it if the cursor is elsewhere.  
	 */
	private void createPrompt() {
		
		textControl.setFont(promptFont);
		document.set(prompt);

		focusListener = new FocusListener() {

			public void focusGained(FocusEvent e) {
				if (document.get().equals(prompt)) {
					textControl.setFont(textFont);
					document.set("");
				}
			}

			public void focusLost(FocusEvent e) {
				if (document.getLength() == 0) {
					textControl.setFont(promptFont);
					document.set(prompt);
				}
			}
			
		};
		textControl.addFocusListener(focusListener);
		textControl.addDisposeListener(new DisposeListener() {

			
			public void widgetDisposed(DisposeEvent e) {
				focusListener = null;
			}
			
		});
	}
	
	/**
	 * Removes focus listener from <code>textControl</code>, turning off
	 * the showing of <code>prompt</code>.
	 */
	private void removePrompt() {
		if (focusListener != null) {
			textControl.removeFocusListener(focusListener);
			focusListener = null;
		}
	}
	
	/* (non-Javadoc)
	 * @see org.eclipse.jface.text.IDocumentListener#documentAboutToBeChanged(org.eclipse.jface.text.DocumentEvent)
	 */
	public void documentAboutToBeChanged(DocumentEvent event) {}

	/* (non-Javadoc)
	 * @see org.eclipse.jface.text.IDocumentListener#documentChanged(org.eclipse.jface.text.DocumentEvent)
	 */
	public void documentChanged(DocumentEvent event) {
		
		if (document.get().equals(prompt)) {
			return;
		}
		
		if (document.getLength() == 0) {
			createPrompt();
		} else {
			removePrompt();
		}
	}
}