/**
 * Copyright (C) 2007 EDIT
 * European Distributed Institute of Taxonomy 
 * http://www.e-taxonomy.eu
 * 
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * See LICENSE.TXT at the top of this package for the full license terms.
 */

package eu.etaxonomy.taxeditor.navigation;

import java.util.Set;

import org.apache.log4j.Logger;
import org.eclipse.core.commands.operations.IUndoContext;
import org.eclipse.core.commands.operations.IUndoableOperation;
import org.eclipse.core.databinding.observable.set.IObservableSet;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IMenuListener;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.action.Separator;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.viewers.CellEditor;
import org.eclipse.jface.viewers.ColumnViewerToolTipSupport;
import org.eclipse.jface.viewers.DoubleClickEvent;
import org.eclipse.jface.viewers.ICellEditorListener;
import org.eclipse.jface.viewers.ICellModifier;
import org.eclipse.jface.viewers.IDoubleClickListener;
import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.TextCellEditor;
import org.eclipse.jface.viewers.TreePath;
import org.eclipse.jface.viewers.TreeSelection;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.jface.viewers.ViewerComparator;
import org.eclipse.swt.SWT;
import org.eclipse.swt.dnd.DND;
import org.eclipse.swt.dnd.DragSource;
import org.eclipse.swt.dnd.DragSourceAdapter;
import org.eclipse.swt.dnd.DragSourceEvent;
import org.eclipse.swt.dnd.DropTarget;
import org.eclipse.swt.dnd.DropTargetAdapter;
import org.eclipse.swt.dnd.DropTargetEvent;
import org.eclipse.swt.dnd.Transfer;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.Tree;
import org.eclipse.swt.widgets.TreeItem;

import eu.etaxonomy.cdm.model.name.NonViralName;
import eu.etaxonomy.cdm.model.reference.ReferenceBase;
import eu.etaxonomy.cdm.model.taxon.Taxon;
import eu.etaxonomy.taxeditor.ITaxEditorConstants;
import eu.etaxonomy.taxeditor.TaxEditorPlugin;
import eu.etaxonomy.taxeditor.actions.TaxonTransfer;
import eu.etaxonomy.taxeditor.actions.cdm.DeleteTaxonAction;
import eu.etaxonomy.taxeditor.actions.ui.OpenNewTaxonEditorAction;
import eu.etaxonomy.taxeditor.actions.ui.OpenTaxonEditorAction;
import eu.etaxonomy.taxeditor.controller.EditorController;
import eu.etaxonomy.taxeditor.controller.GlobalController;
import eu.etaxonomy.taxeditor.controller.PreferencesController;
import eu.etaxonomy.taxeditor.editor.name.CdmParserController;
import eu.etaxonomy.taxeditor.model.CdmSessionDataRepository;
import eu.etaxonomy.taxeditor.model.CdmUtil;
import eu.etaxonomy.taxeditor.model.ICdmTaxonSetListener;
import eu.etaxonomy.taxeditor.operations.MoveTaxonOperation;

/**
 * Taxon tree viewer which responds to events within individual taxa.
 * 
 * Good overview of TreeViewer:
 * http://www.eclipse.org/articles/Article-TreeViewer/TreeViewerArticle.htm
 * 
 * @author p.ciardelli
 * @created 06.05.2008
 * @version 1.0
 */
public class TaxonomicTreeViewer extends TreeViewer{
	private static final Logger logger = Logger
			.getLogger(TaxonomicTreeViewer.class);

	private Tree tree;
	
	private IUndoContext undoContext;

	private boolean initialized = false;

	/**
	 * Creates a lazy-loading taxonomic tree that is sorted by TitleCache, and
	 * which listens for name changes and new taxa
	 * 
	 * @param parent
	 */
	public TaxonomicTreeViewer(Composite parent) {

		// SW.VIRTUAL causes nodes to be loaded on-demand, improving performance
		super(parent, SWT.VIRTUAL);

		undoContext = GlobalController.getWorkbenchUndoContext();
		
		tree = this.getTree();
		tree.setLinesVisible(false);
		tree.setHeaderVisible(false);

		createContent();
		createToolTips();

		initialized = true;
	}

	public boolean isInitialized() {
		return initialized;
	}

	/**
	 * Tool tips are used to show warnings; there is no tooltip mechanism for
	 * individual tree items out-of-the-box, so tipListener below builds one
	 */
	private void createToolTips() {
		ColumnViewerToolTipSupport.enableFor(this);
	}
	
	void createDragAndDrop() {
		final Tree tree = this.getTree();
		Transfer[] types = new Transfer[] { TaxonTransfer.getInstance() };
		int operations = DND.DROP_MOVE | DND.DROP_COPY | DND.DROP_LINK;

		final DragSource source = new DragSource(tree, operations);
		source.setTransfer(types);
		final TreeItem[] dragSourceItem = new TreeItem[1];
		source.addDragListener(new DragSourceAdapter() {
			public void dragStart(DragSourceEvent event) {
				TreeItem[] selection = tree.getSelection();
				if (selection.length > 0) {
					event.doit = true;
					dragSourceItem[0] = selection[0];
				} else {
					event.doit = false;
				}
			}

			public void dragSetData(DragSourceEvent event) {
				Taxon taxon = (Taxon) dragSourceItem[0].getData();
				TaxonTransfer.getInstance().setTaxon(taxon);
			}
		});

		DropTarget target = new DropTarget(tree, operations);
		target.setTransfer(types);
		target.addDropListener(new DropTargetAdapter() {
			public void drop(DropTargetEvent event) {

				// Only drop ONTO another TreeItem OR outside the tree area.
				// Dropping between TreeItems makes no sense because tree is
				// sorted.
				Taxon taxon = (Taxon) dragSourceItem[0].getData();
				Taxon parentTaxon = null;

				// If event.item == null, user has dragged outside of the tree
				if (event.item == null) {
					parentTaxon = null;
				} else {
					parentTaxon = (Taxon) event.item.getData();
					
					// Make sure parentTaxon is not a child
					if (CdmUtil.isTaxonChildOfTaxon(parentTaxon, taxon)) {
						
						MessageDialog.openError(GlobalController.getShell(), "Can't move taxon.",
								"'" + CdmUtil.getDisplayName(taxon) + "' sits above " +
								"'" + CdmUtil.getDisplayName(parentTaxon) + "' " +
								"in the taxonomic hierarchy.");
						
						return;
					}
					
					// Make sure taxon is not being dropped onto itself
					if (taxon.equals(parentTaxon)) {
						return;
					}
				}

				String text = "Move taxon"; //$NON-NLS-1$
				
				IUndoableOperation operation = new MoveTaxonOperation
						(text, undoContext, taxon, parentTaxon);
				
				GlobalController.executeOperation(operation);			
			}
		});
	}

	/**
	 * Taxa with no NameCache - i.e. added with quick name - can be edited in
	 * place in the tax. tree
	 */
	public void createQuickAdd() {

		final Tree tree = this.getTree();
		final TextCellEditor taxTreeNodeEditor = new TextCellEditor(tree);
		
		// Create cell editor for nodes with the QuickNameTaxon singleton
		setCellEditors(new CellEditor[] { taxTreeNodeEditor });
		setColumnProperties(new String[] { "col1" });
		setCellModifier(new ICellModifier() {
			public boolean canModify(Object element, String property) {
				if (element instanceof QuickNameTaxon) {
					return true;
				} else {
					return false;
				}
			}

			public Object getValue(Object element, String property) {
				// If this node is modifiable, name is by definition empty
				return "";
			}

			public void modify(Object element, String property, Object value) {
			}
		});
		
		// If the node loses focus and there is a non zero-length string, create a new taxon
		taxTreeNodeEditor.addListener(new ICellEditorListener() {
			public void applyEditorValue() {
				Object value = taxTreeNodeEditor.getValue();
								
				// Only focus on parent if no taxon is created
				boolean focusParent = false;
				
				if (value instanceof String) {
					String trimmedValue = ((String) value).trim();

					if (trimmedValue.length() > 0) {

						// TODO move this to a create new name operation
						
						NonViralName name = 
								PreferencesController.getInstanceOfPreferredNameClass();
						CdmParserController.parseFullReference(name, trimmedValue);
						
						ReferenceBase sec = QuickNameTaxon.getInstance().getSec();
						
						Taxon taxon = Taxon.NewInstance(name, sec);
						taxon.setTaxonomicParent(QuickNameTaxon.getInstance().getTaxonomicParent(), null, null);
						
						CdmSessionDataRepository.getDefault().saveTaxon(taxon);
					} else {		
						// User has likely hit carriage return
						focusParent = true;
					}
				}
				removeQuickName(focusParent);	
			}

			public void cancelEditor() {
				removeQuickName(true);
			}

			private void removeQuickName(boolean focusParent) {
				// Set focus to quick name's parent?
				if (focusParent) {
					setSelection(new StructuredSelection(
							QuickNameTaxon.getInstance().getTaxonomicParent()), true);
				}
				
				// Clear singleton's parent
				QuickNameTaxon.getInstance().setParent(null);
				
				// Remove quick name from tree
				remove(QuickNameTaxon.getInstance());
			}

			public void editorValueChanged(boolean oldValidState,
					boolean newValidState) {
			}
		});		
	}

	/**
	 * On double click, open name editor
	 */
	public void createDoubleClickListener() {
		this.addDoubleClickListener(new IDoubleClickListener() {

			public void doubleClick(DoubleClickEvent event) {
				if (event.getSelection() instanceof StructuredSelection) {
					Object element = ((StructuredSelection) event
							.getSelection()).getFirstElement();
					if (element instanceof Taxon) {
						EditorController.open((Taxon) element);
					}
				}
			}
		});
	}

	/**
	 * Set up content providers and viewer input
	 */
	private void createContent() {
		
		// Custom content provider
		ITreeContentProvider viewerContentProviderList = new TaxonomicTreeContentProvider();
		setContentProvider(viewerContentProviderList);

		// Label provider that listens for changes to name cache
		IObservableSet observableTaxonSet = 
				CdmSessionDataRepository.getDefault().getObservableTaxa();
		setLabelProvider(new TaxonomicTreeLabelProvider(
						observableTaxonSet));

		// TaxonTreeList added to every time a node is opened with its
		// children, or when a new taxon is added
		setInput(CdmSessionDataRepository.getDefault().getRootTaxa());
				
		// Sort according to "getColumnText" above, i.e. by NameCache
		setComparator(new ViewerComparator());
				
		// Listen for changes to taxa in session data repository
		CdmSessionDataRepository.getDefault().
				addTaxonSetListener(new ICdmTaxonSetListener() {

			
			public void taxaAdded(Set<Taxon> taxa) {
				Taxon lastTaxon = null;
				for (Taxon taxon : taxa) {
					addTaxonToTree(taxon, taxon.getTaxonomicParent());
					lastTaxon = taxon;
				}
				
				// Show last taxon added
				expandToLevel(lastTaxon, 1);
			}

			
			public void taxaRemoved(Set<Taxon> taxa) {
				for (Taxon taxon : taxa) {
					remove(taxon);
				}
			}

			
			public void taxonMoved(Taxon taxon, Taxon newParentTaxon) {
				
				// Save any selection and expansion info ...
				// NOTE: may be superfluous
//				Object[] viewerExpandedElements = null;
//				ISelection viewerSelection = null;
//				viewerExpandedElements = getExpandedElements();
//				viewerSelection = getSelection();
								
				remove(taxon);
				addTaxonToTree(taxon, newParentTaxon);
				
				// ... to restore after taxon has been removed then re-added
//				setExpandedElements(viewerExpandedElements);
//				setSelection(viewerSelection, true);
			}		
		});
	}

	private void addTaxonToTree(Taxon taxon, Taxon parentTaxon) {
		if (parentTaxon == null) {
			add(TreePath.EMPTY, taxon);
			
			// New root effectively changes the viewer's input, theerefore refresh
			refresh();
		} else {
			add(parentTaxon, taxon);
		}
		
	}
	
	/**
	 * If taxon node in the tree is hidden, open and select it
	 * 
	 * @param taxon
	 */
	public void revealTaxon(Taxon taxon) {
		if (taxon == null) {
			return;
		}
		this.setSelection(new StructuredSelection(taxon), true);
		this.reveal(taxon);
	}
	
	public void createMenu() {
		final MenuManager menuManager = new MenuManager();
		menuManager.setRemoveAllWhenShown(true);
		final Menu menu = menuManager.createContextMenu(tree);
		tree.setMenu(menu);

		menuManager.addMenuListener(new IMenuListener() {
			public void menuAboutToShow(IMenuManager manager) {

				Object selection = ((TreeSelection) getSelection())
						.getFirstElement();
				if (!(selection instanceof Taxon)) {
					return;
				}
				final Taxon taxon = (Taxon) selection;
				
				// Open new taxon editor window 
				Action openEditorAction = new OpenTaxonEditorAction(taxon);
				manager.add(openEditorAction);

				// Open new editor with new child taxon. The taxon will not appear in the tree until saved
				Action openNewChildEditorAction = new OpenNewTaxonEditorAction(
						taxon);
				manager.add(openNewChildEditorAction);

				// Create a new node in the tree to enter taxon name directly in the tree.
				manager.add(new QuickNameAction(taxon));
				
				manager.add(new Separator());
				
				// Remove taxon from tree
				Action deleteTaxonAction = new DeleteTaxonAction(taxon);
				manager.add(deleteTaxonAction);
			}
		});
	}
	

	/**
	 * Opens a child node for the parent taxon using the quick name singleton
	 */
	class QuickNameAction extends Action {
		
		private static final String text = "Add child taxon with quick name";
		private final ImageDescriptor image = TaxEditorPlugin.getDefault()
				.getImageDescriptor(ITaxEditorConstants.QUICK_ADD_ICON);
		
		private Taxon parentTaxon;	
		
		QuickNameAction(Taxon parentTaxon) {
			super(text);
			setImageDescriptor(image);
			
			this.parentTaxon = parentTaxon;
		}
		
		public void run() {
			
			QuickNameTaxon quickNameTaxon = QuickNameTaxon.getInstance();
			quickNameTaxon.setParent(parentTaxon);
			
			add(parentTaxon, quickNameTaxon);
			reveal(quickNameTaxon);
			editElement(quickNameTaxon, 0);
		}
	}
}