// $Id$
/**
* Copyright (C) 2007 EDIT
* European Distributed Institute of Taxonomy 
* http://www.e-taxonomy.eu
* 
* The contents of this file are subject to the Mozilla Public License Version 1.1
* See LICENSE.TXT at the top of this package for the full license terms.
*/

package eu.etaxonomy.taxeditor.bulkeditor.referencingobjects;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import org.apache.log4j.Logger;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Table;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.part.ViewPart;
import org.eclipse.ui.progress.IProgressService;

import eu.etaxonomy.cdm.model.common.CdmBase;
import eu.etaxonomy.taxeditor.annotatedlineeditor.LineSelection;
import eu.etaxonomy.taxeditor.model.DescriptionHelper;
import eu.etaxonomy.taxeditor.propertysheet.AgentBasePropertySource;
import eu.etaxonomy.taxeditor.propertysheet.name.NonViralNamePropertySource;
import eu.etaxonomy.taxeditor.propertysheet.reference.ReferencePropertySource;
import eu.etaxonomy.taxeditor.store.CdmStore;

/**
 * @author p.ciardelli
 * @created 08.07.2009
 * @version 1.0
 */
public class ReferencingObjectsView extends ViewPart {
	private static final Logger logger = Logger
			.getLogger(ReferencingObjectsView.class);

	public static final String ID = "taxeditor-bulkeditor.referencingobjectsview";

	private List<CdmBase> referencingObjects;
	private CdmBase referencedObject;
	
	private TableViewer referencingObjectsViewer;

	
	public ReferencingObjectsView() {}
	
	/* (non-Javadoc)
	 * @see org.eclipse.ui.part.WorkbenchPart#createPartControl(org.eclipse.swt.widgets.Composite)
	 */
	@Override
	public void createPartControl(Composite parent) {
		createReferencingObjectsViewer(parent);
	}

    /**
     * @param parent
     */
    private void createReferencingObjectsViewer(Composite parent) {
        referencingObjectsViewer = new TableViewer(parent, SWT.MULTI | SWT.H_SCROLL
									| SWT.V_SCROLL | SWT.FULL_SELECTION);
        createColumns(referencingObjectsViewer);
        referencingObjectsViewer.setContentProvider(new ReferencingObjectsContentProvider());
        referencingObjectsViewer.setLabelProvider(new ReferencingObjectsLabelProvider());
        referencingObjectsViewer.setInput(new ArrayList<CdmBase>());

    }
	
	// This will create the columns for the table
	private void createColumns(TableViewer viewer) {
		Table table = viewer.getTable();
		String[] titles = {"Class", "Description" };
		int[] bounds = { 100, 200 };

		for (int i = 0; i < titles.length; i++) {
			TableViewerColumn column = new TableViewerColumn(viewer, SWT.NONE);
			column.getColumn().setText(titles[i]);
			column.getColumn().setWidth(bounds[i]);
			column.getColumn().setResizable(true);
			column.getColumn().setMoveable(true);
		}
		table.setHeaderVisible(true);
		table.setLinesVisible(true);
	}
    
	/* (non-Javadoc)
	 * @see org.eclipse.ui.part.WorkbenchPart#setFocus()
	 */
	@Override
	public void setFocus() {
		logger.warn("set focus");
		referencingObjectsViewer.getControl().setFocus();
	}

	/**
	 * 
	 */
	public void updateReferencingObjects() {
        IEditorPart editor = getSite().getPage().getActiveEditor();
        ISelection selection = editor.getEditorSite().getSelectionProvider()
                                     .getSelection();

        if ((selection != null) && selection instanceof LineSelection) {
        	final Object entity = ((IStructuredSelection) selection).getFirstElement();
		
			IWorkbench wb = PlatformUI.getWorkbench();
			IProgressService ps = wb.getProgressService();
			try {
				ps.busyCursorWhile(new IRunnableWithProgress() {
					public void run(IProgressMonitor pm) {
						doUpdateReferencingObjects(entity);
					}
				});
			} catch (InvocationTargetException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
        }
		updateView();
	}

	private void doUpdateReferencingObjects(Object entity) {
		CdmStore.NewTransactionalConversation().bind();
        try {
        	referencedObject = getReferencedObject(entity);
        	if (referencedObject != null) {
        		referencingObjects = new ArrayList<CdmBase>(CdmStore.getCommonService().getReferencingObjects(referencedObject));
        		Collections.sort(referencingObjects, new ReferencingClassComparator());
        	}
        } catch (Exception e) {
            logger.error("Error retrieving referencing objects", e);
        }
	}
	
	class ReferencingClassComparator implements Comparator<CdmBase> {

		/* (non-Javadoc)
		 * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
		 */
		public int compare(CdmBase o1, CdmBase o2) {
			String string1 = o1.getClass().getSimpleName();
			String string2 = o2.getClass().getSimpleName();
			int classCompare = string1.compareToIgnoreCase(string2);
			if (classCompare == 0) {
				string1 = DescriptionHelper.getObjectDescription(o1);
				string2 = DescriptionHelper.getObjectDescription(o2);
				return string1.compareToIgnoreCase(string2);
			} else {
				return classCompare;
			}
		}
		
	}
	
	/**
	 * 
	 */
	private void updateView() {
		referencingObjectsViewer.setInput(referencingObjects);
		setContentDescription("'" + referencedObject + "' is referenced by:");
	}
	
	public CdmBase getReferencedObject(Object entity) {
		if (entity instanceof ReferencePropertySource) {
			return ((ReferencePropertySource) entity).getReference();
		}
		if (entity instanceof NonViralNamePropertySource) {
			return ((NonViralNamePropertySource) entity).getName();
		}
		if (entity instanceof AgentBasePropertySource) {
			return ((AgentBasePropertySource) entity).getAgentBase();
		}
		return null;
	}
}
