/**
* Copyright (C) 2007 EDIT
* European Distributed Institute of Taxonomy 
* http://www.e-taxonomy.eu
* 
* The contents of this file are subject to the Mozilla Public License Version 1.1
* See LICENSE.TXT at the top of this package for the full license terms.
*/

package eu.etaxonomy.taxeditor.propertysheet.name;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.Set;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.eclipse.ui.views.properties.ComboBoxPropertyDescriptor;
import org.eclipse.ui.views.properties.IPropertyDescriptor;
import org.eclipse.ui.views.properties.PropertyDescriptor;

import eu.etaxonomy.cdm.model.common.CdmBase;
import eu.etaxonomy.cdm.model.name.BotanicalName;
import eu.etaxonomy.cdm.model.name.NonViralName;
import eu.etaxonomy.cdm.model.name.Rank;
import eu.etaxonomy.cdm.model.name.TaxonNameBase;
import eu.etaxonomy.cdm.model.name.ZoologicalName;
import eu.etaxonomy.cdm.model.reference.ReferenceBase;
import eu.etaxonomy.cdm.model.taxon.TaxonBase;
import eu.etaxonomy.taxeditor.model.NameHelper;
import eu.etaxonomy.taxeditor.propertysheet.AnnotationPropertySource;
import eu.etaxonomy.taxeditor.propertysheet.AnnotationsPropertyDescriptor;
import eu.etaxonomy.taxeditor.propertysheet.ExtensionPropertySource;
import eu.etaxonomy.taxeditor.propertysheet.ExtensionsPropertyDescriptor;
import eu.etaxonomy.taxeditor.propertysheet.ICdmBasePropertySource;
import eu.etaxonomy.taxeditor.propertysheet.MarkerPropertySource;
import eu.etaxonomy.taxeditor.propertysheet.MarkersPropertyDescriptor;
import eu.etaxonomy.taxeditor.propertysheet.reference.ReferencePropertySource;
import eu.etaxonomy.taxeditor.propertysheet.reference.ReferenceSearchDescriptor;
import eu.etaxonomy.taxeditor.propertysheet.type.TypeCollectionPropertySource;
import eu.etaxonomy.taxeditor.propertysheet.type.TypePropertyDescriptor;

/**
 * @author p.ciardelli
 * @created 11.11.2008
 * @version 1.0
 */
public class TaxonBasePropertySource implements ICdmBasePropertySource {
	private static final Logger logger = Logger
			.getLogger(TaxonBasePropertySource.class);
	
	private TaxonBase<?> taxonBase;

    // Property unique keys
	public static final String P_ID_TAXONNAME = "P_ID_TAXONNAME";
	public static final String P_ID_TAXONSEC = "P_ID_TAXONSEC";
	public static final String P_ID_TYPES = "P_ID_TYPES";	
	public static final String P_ID_DOUBTFUL = "P_ID_DOUBTFUL";
	public static final String P_ID_TAXON_MARKERS = "P_ID_TAXON_MARKERS";
	public static final String P_ID_EXTENSIONS = "P_ID_EXTENSIONS";
	public static final String P_ID_TAXON_ANNOTATIONS = "P_ID_TAXON_ANNOTATIONS";
	
    // Property display keys
	public String P_TAXONNAME;
	public static final String P_TAXONSEC = "Secundum";
	public String P_TYPES = "Name Types";
	public static final String P_DOUBTFUL = "Doubtful?";
	public static final String P_TAXON_MARKERS = "Markers";
	public static final String P_EXTENSIONS = "Extensions";
	public static final String P_TAXON_ANNOTATIONS = "Annotations";

	private static final int NOT_DOUBTFUL = 0;
	private static final int DOUBTFUL = 1;	
	
	public TaxonBasePropertySource(TaxonBase<?> taxon, String nameTitle) {
		this.taxonBase = taxon;
		
		this.P_TAXONNAME = nameTitle;
		
		if (taxon != null && taxon.getName() != null && 
				!NameHelper.isNameSupraSpecific(taxon.getName())) {
			P_TYPES = "Specimen Types";
		}
		
		addDescriptor(P_ID_TAXONNAME);
		addDescriptor(P_ID_TAXONSEC);
		addDescriptor(P_ID_TYPES);
		addDescriptor(P_ID_DOUBTFUL);	
		addDescriptor(P_ID_TAXON_ANNOTATIONS);
		addDescriptor(P_ID_TAXON_MARKERS);		
		addDescriptor(P_ID_EXTENSIONS);		
	}

	protected Vector<PropertyDescriptor> descriptors = new Vector<PropertyDescriptor>();
	
	protected void addDescriptor(String id) {
		if (id.equals(P_ID_TAXONNAME)) {
			descriptors.addElement(
					new PropertyDescriptor(P_ID_TAXONNAME, P_TAXONNAME));
		}		
		if (id.equals(P_ID_TAXONSEC)) {
			descriptors.addElement(new ReferenceSearchDescriptor(P_ID_TAXONSEC, P_TAXONSEC, taxonBase.getSec()) {
				protected void saveReference(ReferenceBase reference) {
					setPropertyValue(P_ID_TAXONSEC, reference);
				}
			});
		}
		if (id.equals(P_ID_TYPES)) {
			descriptors.addElement(
				new TypePropertyDescriptor(P_ID_TYPES, P_TYPES, taxonBase.getName()) {
					protected void saveTypes(Set set) {
						setPropertyValue(P_ID_TYPES, set);
					}
				}
			);
		};
		
		if (id.equals(P_ID_DOUBTFUL)) {
			descriptors.addElement(
					new ComboBoxPropertyDescriptor(P_ID_DOUBTFUL, P_DOUBTFUL, 
							new String[] {"no", "yes"}));
		}
		
		if (id.equals(P_ID_TAXON_MARKERS)) {
			descriptors.addElement(
					new MarkersPropertyDescriptor(P_ID_TAXON_MARKERS, P_TAXON_MARKERS, taxonBase));			
		}

		if (id.equals(P_ID_EXTENSIONS)) {
			descriptors.addElement(
					new ExtensionsPropertyDescriptor(P_ID_EXTENSIONS, P_EXTENSIONS, taxonBase));			
		}
		
		// Annotations, listed in custom property descriptor
		if (id.equals(P_ID_TAXON_ANNOTATIONS)) {
			descriptors.addElement(
					new AnnotationsPropertyDescriptor(P_ID_TAXON_ANNOTATIONS, P_TAXON_ANNOTATIONS, taxonBase) {
						protected void saveAnnotations(Set set) {
							setPropertyValue(P_ID_TAXON_ANNOTATIONS, set);
						}
					}
			);
		};
	}
	
	/* (non-Javadoc)
	 * @see org.eclipse.ui.views.properties.IPropertySource#getEditableValue()
	 */
	public Object getEditableValue() {
		return null;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.views.properties.IPropertySource#getPropertyDescriptors()
	 */
	public IPropertyDescriptor[] getPropertyDescriptors() {
		return (IPropertyDescriptor[]) descriptors.toArray(
                new IPropertyDescriptor[descriptors.size()]);
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.views.properties.IPropertySource#getPropertyValue(java.lang.Object)
	 */
	public Object getPropertyValue(Object id) {

		// Edit taxon's name
		if (id.equals(P_ID_TAXONNAME)) {
			
			if (taxonBase == null) {
				logger.warn("no taxon");
				return null;
			}
			TaxonNameBase<?, ?> name = taxonBase.getName();
			
			// Create taxon name as necessary
			if (name == null) {
				name = NonViralName.NewInstance(Rank.SPECIES());
			}
			name = CdmBase.deproxy(name, TaxonNameBase.class);
			// Send taxon name to appropriate property source for submenu
			if (name instanceof BotanicalName) {
				return new BotanicalNamePropertySource((BotanicalName) name);
			}
			if (name instanceof ZoologicalName) {
				return new ZoologicalNamePropertySource((ZoologicalName) name);
			}
			if (name instanceof NonViralName) {
				return new NonViralNamePropertySource((NonViralName<?>) name);
			}
			
		}		
		
		// Edit taxon's sec. reference
		if (id.equals(P_ID_TAXONSEC)) {

			if (taxonBase == null) {
				return null;
			}
			ReferenceBase<?> sec = taxonBase.getSec();
			
			// Create property source for submenu
			ReferencePropertySource secPropertySource = new ReferencePropertySource(sec);
			
			// Add listener to notify taxon of all changes to sec
			secPropertySource.addPropertyChangeListener(new PropertyChangeListener() {
				public void propertyChange(PropertyChangeEvent evt) {
					if (evt.getNewValue() instanceof ReferenceBase) {	
						taxonBase.setSec((ReferenceBase<?>) evt.getNewValue());
					}
				}
			});
			return secPropertySource;
		}
		
		if (id.equals(P_ID_TYPES)) {
			if (taxonBase.getName() != null) {
				// TODO return NameTypeDesignations
				TaxonNameBase<?, ?> name = taxonBase.getName();
				Set<?> typeDesignations = null;
				if (NameHelper.isNameSupraSpecific(name)) {
					typeDesignations = name.getNameTypeDesignations();
				} else {
					typeDesignations = name.getSpecimenTypeDesignations();
				}
				return new TypeCollectionPropertySource(name, typeDesignations);
			}
		}
		
		// Edit doubtful flag
		if (id.equals(P_ID_DOUBTFUL)) {
			if (taxonBase.isDoubtful()) {
				return DOUBTFUL;
			} else {
				return NOT_DOUBTFUL;
			}
		}
		
		if (id.equals(P_ID_TAXON_MARKERS)) {
			return new MarkerPropertySource(taxonBase);
		}

		if (id.equals(P_ID_EXTENSIONS)) {
			return new ExtensionPropertySource(taxonBase);
		}
		        
        if (id.equals(P_ID_TAXON_ANNOTATIONS)) {
			return new AnnotationPropertySource(taxonBase);
        }
		
		return null;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.views.properties.IPropertySource#isPropertySet(java.lang.Object)
	 */
	public boolean isPropertySet(Object id) {
		return false;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.views.properties.IPropertySource#resetPropertyValue(java.lang.Object)
	 */
	public void resetPropertyValue(Object id) {}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.views.properties.IPropertySource#setPropertyValue(java.lang.Object, java.lang.Object)
	 */
	public void setPropertyValue(Object id, Object value) {
        if (id.equals(P_ID_TAXONSEC)) {
        	if (value instanceof ReferenceBase) {
        		taxonBase.setSec((ReferenceBase<?>) value);
        	}
        }
        
        if (id.equals(P_ID_DOUBTFUL)) {
			if (((Integer) value).intValue() == DOUBTFUL) {
				taxonBase.setDoubtful(true);
			} else {
				taxonBase.setDoubtful(false);
			}
        }
	}

	/**
	 * @return the taxonBase
	 */
	public TaxonBase<?> getTaxonBase() {
		return taxonBase;
	}
}
