/**
* Copyright (C) 2007 EDIT
* European Distributed Institute of Taxonomy 
* http://www.e-taxonomy.eu
* 
* The contents of this file are subject to the Mozilla Public License Version 1.1
* See LICENSE.TXT at the top of this package for the full license terms.
*/

package eu.etaxonomy.taxeditor.propertysheet.reference;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.views.properties.ComboBoxPropertyDescriptor;
import org.eclipse.ui.views.properties.IPropertyDescriptor;
import org.eclipse.ui.views.properties.PropertyDescriptor;
import org.eclipse.ui.views.properties.TextPropertyDescriptor;

import eu.etaxonomy.cdm.common.CdmUtils;
import eu.etaxonomy.cdm.hibernate.HibernateProxyHelper;
import eu.etaxonomy.cdm.model.agent.Institution;
import eu.etaxonomy.cdm.model.agent.Team;
import eu.etaxonomy.cdm.model.agent.TeamOrPersonBase;
import eu.etaxonomy.cdm.model.common.TimePeriod;
import eu.etaxonomy.cdm.model.reference.Article;
import eu.etaxonomy.cdm.model.reference.Book;
import eu.etaxonomy.cdm.model.reference.BookSection;
import eu.etaxonomy.cdm.model.reference.CdDvd;
import eu.etaxonomy.cdm.model.reference.Database;
import eu.etaxonomy.cdm.model.reference.Generic;
import eu.etaxonomy.cdm.model.reference.InProceedings;
import eu.etaxonomy.cdm.model.reference.Journal;
import eu.etaxonomy.cdm.model.reference.Map;
import eu.etaxonomy.cdm.model.reference.Patent;
import eu.etaxonomy.cdm.model.reference.PersonalCommunication;
import eu.etaxonomy.cdm.model.reference.PrintSeries;
import eu.etaxonomy.cdm.model.reference.PrintedUnitBase;
import eu.etaxonomy.cdm.model.reference.Proceedings;
import eu.etaxonomy.cdm.model.reference.PublicationBase;
import eu.etaxonomy.cdm.model.reference.ReferenceBase;
import eu.etaxonomy.cdm.model.reference.ReferenceType;
import eu.etaxonomy.cdm.model.reference.Report;
import eu.etaxonomy.cdm.model.reference.SectionBase;
import eu.etaxonomy.cdm.model.reference.Thesis;
import eu.etaxonomy.cdm.model.reference.WebPage;
import eu.etaxonomy.taxeditor.editor.MultiPageTaxonEditor;
import eu.etaxonomy.taxeditor.model.Resources;
import eu.etaxonomy.taxeditor.model.TimeHelper;
import eu.etaxonomy.taxeditor.propertysheet.ICdmBasePropertySource;
import eu.etaxonomy.taxeditor.propertysheet.MarkerPropertySource;
import eu.etaxonomy.taxeditor.propertysheet.MarkersPropertyDescriptor;
import eu.etaxonomy.taxeditor.propertysheet.TimePeriodPropertySource;
import eu.etaxonomy.taxeditor.propertysheet.YearValidator;

/**
 * Supplies the properties of a <code>ReferenceBase</code> object for display by
 * a <code>PropertySheet</code>.
 * <p>
 * Includes a drop-down menu that allows the user to change <code>ReferenceBase</code> 
 * subclasses. This causes all fields except <code>TitleCache</code> to be wiped clean.
 * <p>
 * Normally, this would be called from another <code>IPropertySource</code>  
 * as follows:
 * <pre>
 * public Object getPropertyValue(Object id) {
 * 
 * 		...
 * 
 * 		if (id.equals(P_ID_BIBREF)) {
 *			ReferenceBase citation = descriptionElement.getCitation();
 *			ReferencePropertySource bibRefPropertySource = new ReferencePropertySource(citation);
 *			bibRefPropertySource.addPropertyChangeListener(new PropertyChangeListener() {
 *				@Override
 *				public void propertyChange(PropertyChangeEvent evt) {
 *					descriptionElement.setCitation((ReferenceBase) evt.getNewValue());
 *				}
 *			});
 *			return bibRefPropertySource;
 *		}		
 * </pre>
 * <p>
 * Note: if <code>P_ID_BIBREF</code> in the above example is associated with an editable 
 * <code>TextPropertyDescriptor</code>, it will initalize its cell editor by calling 
 * <code>getEditableValue()</code>. If associated with a non-editable <code>PropertyDescriptor</code>, 
 * it will get its value from <code>getString()</code>.
 * @author p.ciardelli
 * @created 03.11.2008
 * @version 1.0
 */
public class ReferencePropertySource implements ICdmBasePropertySource {
	private static final Logger logger = Logger
			.getLogger(ReferencePropertySource.class);

	ReferenceBase<?> reference;

	private ReferenceType referenceType;
	
    // Property unique keys
	public static final String P_ID_SEARCH = "P_ID_SEARCH";
	public static final String P_ID_EDITABLECACHE = "P_ID_EDITABLECACHE";	
	public static final String P_ID_REFERENCETYPE = "P_ID_REFERENCETYPE";
	public static final String P_ID_AUTHORTEAM = "P_ID_AUTHORTEAM";
	public static final String P_ID_URI = "P_ID_URI";
	public static final String P_ID_DATEPUBLISHED = "P_ID_DATEPUBLISHED";
	public static final String P_ID_TITLE = "P_ID_TITLE";
	public static final String P_ID_PAGES = "P_ID_PAGES";
	public static final String P_ID_SERIES = "P_ID_SERIES";
	public static final String P_ID_VOLUME = "P_ID_VOLUME";
	public static final String P_ID_EDITOR = "P_ID_EDITOR";
	public static final String P_ID_PLACEPUBLISHED = "P_ID_PLACEPUBLISHED";
	public static final String P_ID_PUBLISHER = "P_ID_PUBLISHER";
	public static final String P_ID_ISSN = "P_ID_ISSN";
	public static final String P_ID_INSTITUTION = "P_ID_INSTITUTION";
	public static final String P_ID_SCHOOL = "P_ID_SCHOOL";
	public static final String P_ID_SERIESPART = "P_ID_SERIESPART";
	public static final String P_ID_ISBN = "P_ID_ISBN";
	public static final String P_ID_ORGANIZATION = "P_ID_ORGANIZATION";
	public static final String P_ID_JOURNAL = "P_ID_JOURNAL";
	public static final String P_ID_BOOKTITLE = "P_ID_BOOKTITLE";
	public static final String P_ID_CHAPTER = "P_ID_CHAPTER";
	public static final String P_ID_EDITION = "P_ID_EDITION";
	public static final String P_ID_NUMBER = "P_ID_NUMBER";
	public static final String P_ID_ANNOTE = "P_ID_ANNOTE";
	public static final String P_ID_ADDRESS = "P_ID_ADDRESS";
	public static final String P_ID_HOWPUBLISHED = "P_ID_HOWPUBLISHED";
	public static final String P_ID_REPORTTYPE = "P_ID_REPORTTYPE";
	public static final String P_ID_MONTH = "P_ID_MONTH";
	public static final String P_ID_EPRINT = "P_ID_EPRINT";
	public static final String P_ID_NOTE = "P_ID_NOTE";
	public static final String P_ID_CROSSREF = "P_ID_CROSSREF";
	public static final String P_ID_PROTECT_CACHE = "P_ID_PROTECT_CACHE";
	public static final String P_ID_ABSTRACT = "P_ID_ABSTRACT";
	public static final String P_ID_INREFERENCE = "P_ID_INREFERENCE";
	public static final String P_ID_REFERENCE_MARKERS = "P_ID_REFERENCE_MARKERS";
		
    // Property display keys
	public static final String P_SEARCH = "Search";
	public static final String P_EDITABLECACHE = "Editable Cache";
	public static final String P_PROTECT_CACHE = "Protect Cache from overwriting?";
	public static final String P_REFERENCETYPE = "Reference Type";
	public static final String P_AUTHORTEAM = "Author Team (Cache)";
	public static final String P_DATEPUBLISHED = "Date Published";
	public static final String P_TITLE = "Title";
	public static final String P_VOLUME = "Volume";
	public static final String P_EDITION = "Edition";
	public static final String P_SERIES = "Series";
	public static final String P_PAGES = "Pages";
	public static final String P_EDITOR = "Editor";
	public static final String P_PUBLISHER = "Publisher";
	public static final String P_PLACEPUBLISHED = "Place Published";
	public static final String P_ISSN = "ISSN";
	public static final String P_INSTITUTION = "Institution";
	public static final String P_SCHOOL = "School";
	public static final String P_SERIESPART = "Series Part";
	public static final String P_ISBN = "ISBN";
	public static final String P_URI = "URI";
	public static final String P_ORGANIZATION = "Organization";
	public static final String P_JOURNAL = "Journal";
	public static final String P_BOOKTITLE = "Book Title";
	public static final String P_CHAPTER = "Chapter";
	public static final String P_NUMBER = "Number";
	public static final String P_ANNOTE = "Annote";
	public static final String P_ADDRESS = "Address";
	public static final String P_HOWPUBLISHED = "How Published";
	public static final String P_REPORTTYPE = "Report Type";
	public static final String P_MONTH = "Month";
	public static final String P_EPRINT = "E-Print";
	public static final String P_NOTE = "Note";
	public static final String P_ABSTRACT = "Abstract";
	public static final String P_INREFERENCE = "In Reference";
	public static final String P_REFERENCE_MARKERS = "Markers";
	
	protected static HashMap<Class<?>, String> referenceTypeMap = null;

	private static final int CACHE_NOT_PROTECTED = 0;
	private static final int CACHE_PROTECTED = 1;
	
	protected void populateReferenceTypes() {
		
		// LinkedHashMap maintains insertion order
		referenceTypeMap = new LinkedHashMap<Class<?>, String>();
		
		referenceTypeMap.put(ReferenceBase.class, "");
		referenceTypeMap.put(Article.class, "Article");
		referenceTypeMap.put(Generic.class, "Generic");
		referenceTypeMap.put(Patent.class, "Patent");
		referenceTypeMap.put(PersonalCommunication.class, "Personal Communication");
		referenceTypeMap.put(CdDvd.class, "CD / DVD");
		referenceTypeMap.put(Database.class, "Database");
		referenceTypeMap.put(Journal.class, "Journal");
		referenceTypeMap.put(Map.class, "Map");
		referenceTypeMap.put(Book.class, "Book");
		referenceTypeMap.put(Proceedings.class, "Proceedings");
		referenceTypeMap.put(PrintSeries.class, "Print Series");
		referenceTypeMap.put(Report.class, "Report");
		referenceTypeMap.put(Thesis.class, "Thesis");
		referenceTypeMap.put(WebPage.class, "Web Page");
		referenceTypeMap.put(BookSection.class, "Book Section");
		referenceTypeMap.put(InProceedings.class, "In Proceedings");
	}
	
	protected String[] getReferenceTypeArray() {
		if (referenceTypeMap == null) {
			populateReferenceTypes();
		}
		return referenceTypeMap.values().toArray(new String[] {});
	}
	
	protected Set<Class<?>> getReferenceClassSet() {
		return referenceTypeMap.keySet();
	}
		
	public ReferencePropertySource(ReferenceBase<?> reference) {
		this(reference, null);
	}

	/**
	 * @param inReference
	 * @param inReferenceType
	 */
	public ReferencePropertySource(ReferenceBase<?> reference,
			ReferenceType referenceType) {
		
		this.referenceType = referenceType;
		
		// Default type of ReferenceBase is Generic
		if (reference == null) {
			reference = Generic.NewInstance();
		} else {
			reference = HibernateProxyHelper.deproxy(reference, ReferenceBase.class);
		}
		this.reference = reference;

		initDescriptors();
	}

	protected void initDescriptors() {
		
		List<String> displayFields = new ArrayList<String>();
	
		// Drop-down menu to change reference type
		displayFields.add(P_ID_REFERENCETYPE);
		
		// ReferenceBase fields
//		displayFields.add(P_ID_SEARCH);
		displayFields.add(P_ID_EDITABLECACHE);
		displayFields.add(P_ID_PROTECT_CACHE);
		displayFields.add(P_ID_AUTHORTEAM);
		displayFields.add(P_ID_URI);
		
		Class<?> referenceClass = reference.getClass();
			
		displayFields.add(P_ID_DATEPUBLISHED);
		displayFields.add(P_ID_TITLE);
	
		if (referenceClass == Article.class) {
			displayFields.add(P_ID_INREFERENCE);
			displayFields.add(P_ID_PAGES);
			displayFields.add(P_ID_SERIES);
			displayFields.add(P_ID_VOLUME);
		}
	
		if (referenceClass == Generic.class) {
			displayFields.add(P_ID_PAGES);
			displayFields.add(P_ID_SERIES);
			displayFields.add(P_ID_VOLUME);
			displayFields.add(P_ID_EDITOR);
		}			
			
		if (referenceClass == Patent.class) {
			// No additional fields
		}
			
		if (referenceClass == PersonalCommunication.class) {
			// No additional fields				
		}
		if (reference instanceof PublicationBase) {
			displayFields.add(P_ID_PLACEPUBLISHED);
			displayFields.add(P_ID_PUBLISHER);
		}
		
		if (referenceClass == CdDvd.class) {
			// No additional fields							
		}
			
		if (referenceClass == Database.class) {
			// No additional fields					
		}
			
		if (referenceClass == Journal.class) {	
			displayFields.add(P_ID_ISSN);
		}

		if (referenceClass == Map.class) {
			// No additional fields					
		}				
		
		if (referenceClass == PrintSeries.class) {
			displayFields.add(P_ID_SERIES);					
		}
		
		if (referenceClass == Report.class) {
			displayFields.add(P_ID_INSTITUTION);					
		}
		
		if (referenceClass == Thesis.class) {
			displayFields.add(P_ID_SCHOOL);						
		}
		
		if (referenceClass == WebPage.class) {
			// No additional fields		
		}
			
		if (reference instanceof PrintedUnitBase) {
				
			displayFields.add(P_ID_EDITOR);
			displayFields.add(P_ID_INREFERENCE);
			displayFields.add(P_ID_PAGES);
			displayFields.add(P_ID_SERIESPART);
			displayFields.add(P_ID_VOLUME);
				
			if (referenceClass == Book.class) {
				displayFields.add(P_ID_ISBN);
				displayFields.add(P_ID_EDITION);
			}
				
			if (referenceClass == Proceedings.class) {
				displayFields.add(P_ID_ORGANIZATION);							
			}
		}
			
		if (reference instanceof SectionBase) {
				
			displayFields.add(P_ID_PAGES);
				
			if (referenceClass == BookSection.class) {
				displayFields.add(P_ID_INREFERENCE);
			}
				
			if (referenceClass == InProceedings.class) {
				displayFields.add(P_ID_INREFERENCE);					
			}
		}
		
		displayFields.add(P_ID_ABSTRACT);
		displayFields.add(P_ID_REFERENCE_MARKERS);
		
		// There was a problem with duplicate descriptors in the above code
		List<String> displayFieldsTemp = new ArrayList<String>();
		for (String field : displayFields) {
			if (displayFieldsTemp.contains(field)) {
				continue;
			}
			addDescriptor(field);
			displayFieldsTemp.add(field);
		}
	}

	protected Vector<PropertyDescriptor> descriptors = new Vector<PropertyDescriptor>();
	
	protected void addDescriptor(String id) {
		
//		// Reference search
//		if (id.equals(P_ID_SEARCH) && isMultiPageTaxonEditorActive()) {
//			descriptors.addElement(
//					new ReferenceSearchDescriptor(P_ID_SEARCH, P_SEARCH, reference, new ReferenceType[] { referenceType }) {
//						protected void saveReference(ReferenceBase reference) {
//							setPropertyValue(P_ID_SEARCH, reference);
//						}
//			});
//		}
		
		// Editable cache
		if (id.equals(P_ID_EDITABLECACHE)) {
			descriptors.addElement(
					new TextPropertyDescriptor(P_ID_EDITABLECACHE, P_EDITABLECACHE));
		}
		
		// Protect cache?
		if (id.equals(P_ID_PROTECT_CACHE)) {
			descriptors.addElement(
					new ComboBoxPropertyDescriptor(P_ID_PROTECT_CACHE, P_PROTECT_CACHE, 
							new String[] {"no", "yes"}));			
		}
		
		// Reference type
		if (id.equals(P_ID_REFERENCETYPE)) {
			descriptors.addElement(createReferenceTypeDescriptor());
		}
		
		boolean isProtectedCache = reference.isProtectedTitleCache();
		
		// Author team
		if (id.equals(P_ID_AUTHORTEAM)) {
			descriptors.addElement(
					isProtectedCache?
							new PropertyDescriptor(P_ID_AUTHORTEAM, P_AUTHORTEAM) :
							new TextPropertyDescriptor(P_ID_AUTHORTEAM, P_AUTHORTEAM));			
		}
				
		// Date published
		if (id.equals(P_ID_DATEPUBLISHED)) {
			
			// TimePeriodPropertySource returns TimePeriod.getYear() - make this editable
			if (isProtectedCache) {
				descriptors.addElement(new PropertyDescriptor(
						P_ID_DATEPUBLISHED, P_DATEPUBLISHED));
			} else {
				TextPropertyDescriptor yearDescriptor = new TextPropertyDescriptor(
						P_ID_DATEPUBLISHED, P_DATEPUBLISHED);
				yearDescriptor.setValidator(new YearValidator());
				descriptors.addElement(yearDescriptor);
			}
		} 
				
		// URI
		if (id.equals(P_ID_URI)) {
			descriptors.addElement(
					isProtectedCache?
							new PropertyDescriptor(P_ID_URI, P_URI) :
							new TextPropertyDescriptor(P_ID_URI, P_URI));			
		} 
		
		// Title
		if (id.equals(P_ID_TITLE)) {
			descriptors.addElement(
					isProtectedCache?
							new PropertyDescriptor(P_ID_TITLE, P_TITLE) :
							new TextPropertyDescriptor(P_ID_TITLE, P_TITLE));			
		} 
		
		// In reference
		if (id.equals(P_ID_INREFERENCE)) {
			if (isProtectedCache) {
				descriptors.addElement(new PropertyDescriptor(P_ID_INREFERENCE, P_INREFERENCE));
			} else {
				ReferenceBase inRef = null;
				ReferenceType inRefType = null;
				if (reference instanceof Article) {
					inRef = ((Article) reference).getInJournal();
					inRefType = ReferenceType.Journal;
				}
				if (reference instanceof PrintedUnitBase) {
					inRef = ((PrintedUnitBase) reference).getInSeries();
					inRefType = ReferenceType.Journal;
				}
				if (reference instanceof BookSection) {
					inRef = ((BookSection) reference).getInBook();
					inRefType = ReferenceType.Journal;
				}
				if (reference instanceof InProceedings) {
					inRef = ((InProceedings) reference).getInProceedings();
					inRefType = ReferenceType.Journal;
				}
				descriptors.addElement(new ReferenceSearchDescriptor(P_ID_INREFERENCE,P_INREFERENCE, inRef, new ReferenceType[] { inRefType }) {
					protected void saveReference(ReferenceBase reference) {
						setPropertyValue(P_ID_INREFERENCE, reference);
					}
				});	
			}			
		}
		
		// Pages
		if (id.equals(P_ID_PAGES)) {
			descriptors.addElement(
					isProtectedCache?
							new PropertyDescriptor(P_ID_PAGES, P_PAGES) :
							new TextPropertyDescriptor(P_ID_PAGES, P_PAGES));			
		} 
		
		// Series
		if (id.equals(P_ID_SERIES)) {
			descriptors.addElement(
					isProtectedCache?
							new PropertyDescriptor(P_ID_SERIES, P_SERIES) :
							new TextPropertyDescriptor(P_ID_SERIES, P_SERIES));			
		} 
		
		// Volume
		if (id.equals(P_ID_VOLUME )) {
			descriptors.addElement(
					isProtectedCache?
							new PropertyDescriptor(P_ID_VOLUME, P_VOLUME) :
							new TextPropertyDescriptor(P_ID_VOLUME, P_VOLUME));			
		}
		
		// Editor
		if (id.equals(P_ID_EDITOR)) {
			descriptors.addElement(
					isProtectedCache?
							new PropertyDescriptor(P_ID_EDITOR, P_EDITOR) :
							new TextPropertyDescriptor(P_ID_EDITOR, P_EDITOR));			
		} 

		// Edition
		if (id.equals(P_ID_EDITION)) {
			descriptors.addElement(
					isProtectedCache?
							new PropertyDescriptor(P_ID_EDITION, P_EDITION) :
							new TextPropertyDescriptor(P_ID_EDITION, P_EDITION));			
		} 
		
		// Place published
		if (id.equals(P_ID_PLACEPUBLISHED)) {
			descriptors.addElement(
					isProtectedCache?
							new PropertyDescriptor(P_ID_PLACEPUBLISHED, P_PLACEPUBLISHED) :
							new TextPropertyDescriptor(P_ID_PLACEPUBLISHED, P_PLACEPUBLISHED));			
		}
		
		// Publisher
		if (id.equals(P_ID_PUBLISHER)) {
			descriptors.addElement(
					isProtectedCache?
							new PropertyDescriptor(P_ID_PUBLISHER, P_PUBLISHER) :
							new TextPropertyDescriptor(P_ID_PUBLISHER, P_PUBLISHER));			
		} 
		
		// ISSN
		if (id.equals(P_ID_ISSN)) {
			descriptors.addElement(
					isProtectedCache?
							new PropertyDescriptor(P_ID_ISSN, P_ISSN) :
							new TextPropertyDescriptor(P_ID_ISSN, P_ISSN));			
		} 
		
		// Institution
		if (id.equals(P_ID_INSTITUTION)) {
			descriptors.addElement(
					isProtectedCache?
							new PropertyDescriptor(P_ID_INSTITUTION, P_INSTITUTION) :
							new TextPropertyDescriptor(P_ID_INSTITUTION, P_INSTITUTION));			
		} 
		
		// School
		if (id.equals(P_ID_SCHOOL)) {
			descriptors.addElement(
					isProtectedCache?
							new PropertyDescriptor(P_ID_SCHOOL, P_SCHOOL) :
							new TextPropertyDescriptor(P_ID_SCHOOL, P_SCHOOL));			
		} 
				
		// Series part
		if (id.equals(P_ID_SERIESPART)) {
			descriptors.addElement(
					isProtectedCache?
							new PropertyDescriptor(P_ID_SERIESPART, P_SERIESPART) :
							new TextPropertyDescriptor(P_ID_SERIESPART, P_SERIESPART));			
		} 
		
		// ISBN
		if (id.equals(P_ID_ISBN)) {
			descriptors.addElement(
					isProtectedCache?
							new PropertyDescriptor(P_ID_ISBN, P_ISBN) :
							new TextPropertyDescriptor(P_ID_ISBN, P_ISBN));
		}
		
		// Organization
		if (id.equals(P_ID_ORGANIZATION)) {
			descriptors.addElement(
					isProtectedCache?
							new PropertyDescriptor(P_ID_ORGANIZATION, P_ORGANIZATION) :
							new TextPropertyDescriptor(P_ID_ORGANIZATION, P_ORGANIZATION));			
		}
		
		// Abstract
		if (id.equals(P_ID_ABSTRACT)) {
			descriptors.addElement(
					isProtectedCache?
							new PropertyDescriptor(P_ID_ABSTRACT, P_ABSTRACT) :
							new TextPropertyDescriptor(P_ID_ABSTRACT, P_ABSTRACT));			
		}
		
		// Markers
		if (id.equals(P_ID_REFERENCE_MARKERS)) {
			descriptors.addElement(
							new MarkersPropertyDescriptor(P_ID_REFERENCE_MARKERS, P_REFERENCE_MARKERS, reference));
		}
	}

	/**
	 * @return
	 */
	private boolean isMultiPageTaxonEditorActive() {
		IWorkbenchWindow window = PlatformUI.getWorkbench().getActiveWorkbenchWindow();
		if (window == null) {
			return false;
		}
		IEditorPart editor = window.getActivePage().getActiveEditor();
		return (editor instanceof MultiPageTaxonEditor);
	}

	/**
	 * The default implementation of <code>ReferencePropertySource</code> will
	 * search all <code>ReferenceBase</code> classes.
	 * 
	 * @return
	 */
//	protected int getSearchType() {
//		return IReferenceSearch.BIBREF;
//	}

	/**
	 * The default implementation of <code>ReferencePropertySource</code> will show
	 * a drop-down of reference types, but subclasses may want to show a non-editable
	 * text descriptor of the reference type. 
	 */
	protected PropertyDescriptor createReferenceTypeDescriptor() {
		if (referenceType == null) {
			return new ComboBoxPropertyDescriptor(
					P_ID_REFERENCETYPE, P_REFERENCETYPE, getReferenceTypeArray());
		} else {
			return new PropertyDescriptor(P_ID_REFERENCETYPE, P_REFERENCETYPE);
		}
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.views.properties.IPropertySource#getEditableValue()
	 */
	public Object getEditableValue() {
		return CdmUtils.Nz(reference.getTitleCache());
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.views.properties.IPropertySource#getPropertyDescriptors()
	 */
	public IPropertyDescriptor[] getPropertyDescriptors() {
		return (IPropertyDescriptor[]) descriptors.toArray(
                new IPropertyDescriptor[descriptors.size()]);
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.views.properties.IPropertySource#getPropertyValue(java.lang.Object)
	 */
	public Object getPropertyValue(Object id) {
				
		// Reference Type: get integer index of drop-down menu
		if (id.equals(P_ID_REFERENCETYPE)) {
			if (referenceType != null) {
				String message = referenceType.getMessage();
				return message;
			} else {
				Class<?> referenceClass = reference.getClass();
				int menuIndex = 0;
				for (Class<?> clazz : getReferenceClassSet()) {
					if (clazz == referenceClass) {
						return menuIndex;
					}
					menuIndex++;
				}
				logger.warn("Reference class type " + referenceClass + " not in reference type menu.");
				return 0;
			}
		}

		// Editable cache
		if (id.equals(P_ID_EDITABLECACHE)) {
			return (CdmUtils.Nz(reference.getTitleCache()));
		}
		
		// Protect cache?
		if (id.equals(P_ID_PROTECT_CACHE)) {
			if (reference.isProtectedTitleCache()) {
				return CACHE_PROTECTED;
			} else {
				return CACHE_NOT_PROTECTED;
			}
//			return reference.isProtectedTitleCache();
		}
		
		// Author Team: for now, just edit the cache
		// TODO AuthorTeamPropertySource
		if (id.equals(P_ID_AUTHORTEAM)) {
			TeamOrPersonBase<?> authorTeam = reference.getAuthorTeam();
			
			if (authorTeam == null) {
				return "";
			}
			
			return CdmUtils.Nz(authorTeam.getTitleCache());
		}
				
		// URI
		if (id.equals(P_ID_URI)) {
			return CdmUtils.Nz(reference.getUri());
		}
		
		// Date published
		if (id.equals(P_ID_DATEPUBLISHED)) {
			TimePeriod datePublished = (TimePeriod) invokeMethod(reference, "getDatePublished", null);
			
			if (datePublished == null) {
				datePublished = TimePeriod.NewInstance();
			}
			
			if (reference.isProtectedTitleCache()) {
				return CdmUtils.Nz(reference.getYear());
			}
			
			TimePeriodPropertySource timePeriodPropertySource = 
					new TimePeriodPropertySource(datePublished);
			timePeriodPropertySource.addPropertyChangeListener(new PropertyChangeListener() {
				public void propertyChange(PropertyChangeEvent evt) {
					if (evt.getNewValue() instanceof TimePeriod) {	
						reference.setDatePublished((TimePeriod) evt.getNewValue());
					}
				}
			});
			return timePeriodPropertySource;
		}
		
		// Title: used by StrictReferenceBase
		if (id.equals(P_ID_TITLE)) {
			String title = (String) invokeMethod(reference, "getTitle", null);
			return CdmUtils.Nz(title);
		}
		
		// InReference: return an instance of ReferencePropertySource
		if (id.equals(P_ID_INREFERENCE)) {
						
			ReferenceBase<?> inReference = null;
			ReferenceType inReferenceType = null; 
			
			if (reference instanceof Article) {
				inReferenceType = ReferenceType.Journal;
				inReference = ((Article) reference).getInJournal();
				if (inReference == null) {
					inReference = Journal.NewInstance();
					((Article) reference).setInJournal((Journal) inReference);
				}
			}
			if (reference instanceof PrintedUnitBase) {
				inReferenceType = ReferenceType.PrintSeries;
				inReference = ((PrintedUnitBase) reference).getInSeries();
				if (inReference == null) {
					inReference = PrintSeries.NewInstance();
					((PrintedUnitBase) reference).setInSeries((PrintSeries) inReference);
				}				
			}			
			if (reference instanceof BookSection) {
				inReferenceType = ReferenceType.Book;
				inReference = ((BookSection) reference).getInBook();
				if (inReference == null) {
					inReference = Book.NewInstance();
					((BookSection) reference).setInBook((Book) inReference);
				}
			}			
			if (reference instanceof InProceedings) {
				inReferenceType = ReferenceType.Proceedings;
				inReference = ((InProceedings) reference).getInProceedings();
				if (inReference == null) {
					inReference = Proceedings.NewInstance();
					((InProceedings) reference).setInProceedings((Proceedings) inReference);
				}				
			}
			
			if (reference.isProtectedTitleCache()) {
				return inReference.getTitleCache(); 
			}
			
			ReferencePropertySource bibRefPropertySource = new ReferencePropertySource(inReference, inReferenceType);
			bibRefPropertySource.addPropertyChangeListener(new PropertyChangeListener() {
				public void propertyChange(PropertyChangeEvent evt) {
					Object newInRef = evt.getNewValue();
					setInReference(newInRef);
				}
			});
			return bibRefPropertySource;
		}
		
		// Pages
		if (id.equals(P_ID_PAGES)) {
			String pages = (String) invokeMethod(reference, "getPages", null);
			return CdmUtils.Nz(pages);
		}
		
		// Series
		if (id.equals(P_ID_SERIES)) {
			String series = (String) invokeMethod(reference, "getSeries", null);
			return CdmUtils.Nz(series); 
		}

		// Volume
		if (id.equals(P_ID_VOLUME)) {
			String volume = (String) invokeMethod(reference, "getVolume", null);
			return CdmUtils.Nz(volume);			
		}

		// Editor
		if (id.equals(P_ID_EDITOR)) {
			String editor = (String) invokeMethod(reference, "getEditor", null);
			return CdmUtils.Nz(editor);			
		}

		// Edition
		if (id.equals(P_ID_EDITION)) {
			String edition = (String) invokeMethod(reference, "getEdition", null);
			return CdmUtils.Nz(edition);			
		}
		
		// Place published
		if (id.equals(P_ID_PLACEPUBLISHED)) {
			String placePublished = (String) invokeMethod(reference, "getPlacePublished", null);
			return CdmUtils.Nz(placePublished);
		}
		
		// Publisher
		if (id.equals(P_ID_PUBLISHER)) {
			String publisher = (String) invokeMethod(reference, "getPublisher", null);
			return CdmUtils.Nz(publisher);			
		}

		// ISSN
		if (id.equals(P_ID_ISSN)) {
			String issn = (String) invokeMethod(reference, "getIssn", null);
			return CdmUtils.Nz(issn);			
		}

		// Institution
		if (id.equals(P_ID_INSTITUTION)) {
			Institution institution = ((Report) reference).getInstitution();
			return institution == null ? "" : CdmUtils.Nz(institution.getName());			
		}
		
		// ISBN
		if (id.equals(P_ID_ISBN)) {
			String isbn = (String) invokeMethod(reference, "getIsbn", null);
			return CdmUtils.Nz(isbn);				
		}

		// Organization
		if (id.equals(P_ID_ORGANIZATION)) {
			String organization = (String) invokeMethod(reference, "getOrganization", null);
			return CdmUtils.Nz(organization);			
		}

		if (id.equals(P_ID_ABSTRACT)) {
			return CdmUtils.Nz(reference.getReferenceAbstract());
		}
		
		if (id.equals(P_ID_SERIESPART)) {
			return CdmUtils.Nz(((PrintedUnitBase) reference).getSeriesPart());
		}
		
		if (id.equals(P_ID_SCHOOL)) {
			Institution school = ((Thesis) reference).getSchool();
			return school == null ? "" : CdmUtils.Nz(school.getName());	
		}
		
		if (id.equals(P_ID_REFERENCE_MARKERS)) {
			return new MarkerPropertySource(reference);
		}
				
		return "";
	}

	/**
	 * @param newInRef
	 */
	private void setInReference(Object newInRef) {
		if (reference instanceof Article && newInRef instanceof Journal) {
			((Article) reference).setInJournal((Journal) newInRef);
		}
		if (reference instanceof PrintedUnitBase && newInRef instanceof PrintSeries) {
			((PrintedUnitBase) reference).setInSeries((PrintSeries) newInRef);
		}
		if (reference instanceof BookSection &&newInRef instanceof Book) {
			((BookSection) reference).setInBook((Book) newInRef);
		}
		if (reference instanceof InProceedings && newInRef instanceof Proceedings) {
			((InProceedings) reference).setInProceedings((Proceedings) newInRef);
		}
	}

	/**
	 * Executes <code>methodName</code> on <code>obj</code> with
	 * the argument <code>arg</code>.
	 * <p>
	 * On failure, returns <code>null</code>.
	 * <p>
	 * NOTE:  both <code>Class.getMethod()</code> and <code>Method.invoke()</code>
	 * can be called with the <code>Object... args</code> approach, which allows
	 * the user to pass in a varying number of arguments. To keep it simple, and since
	 * this is usually only getters or setters, we are using only one argument.
	 * @param obj
	 * @param methodName
	 * @param args
	 * @return
	 */
	private Object invokeMethod(Object obj, String methodName, Object arg) {
		try {
			Method method = null;
			if (arg == null) {
				method = obj.getClass().getMethod(methodName);
				return method.invoke(obj);
			} else {
				method = obj.getClass().getMethod(methodName, arg.getClass());
				return method.invoke(obj, arg);
			}
		} catch (SecurityException e) {
			logger.error(e.getMessage());
		} catch (IllegalArgumentException e) {
			logger.error(e.getMessage());
		} catch (NoSuchMethodException e) {
			logger.error(e.getMessage());
		} catch (IllegalAccessException e) {
			logger.error(e.getMessage());
		} catch (InvocationTargetException e) {
			logger.error(e.getMessage());
		}
		return null;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.views.properties.IPropertySource#isPropertySet(java.lang.Object)
	 */
	public boolean isPropertySet(Object id) {
		return false;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.views.properties.IPropertySource#resetPropertyValue(java.lang.Object)
	 */
	public void resetPropertyValue(Object id) {}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.views.properties.IPropertySource#setPropertyValue(java.lang.Object, java.lang.Object)
	 */
	public void setPropertyValue(Object id, Object value) {
		
		// Search result
		if (id.equals(P_ID_SEARCH)) {
			if (value instanceof ReferenceBase) {
				reference = (ReferenceBase<?>) value;
			}
		}

		// Protect cache? 
		if (id.equals(P_ID_PROTECT_CACHE)) {
			if (((Integer) value).intValue() == CACHE_PROTECTED) {
				reference.setProtectedTitleCache(true);
			} else {
				reference.setProtectedTitleCache(false);
			}
		}
				
		// Reference Type
		if (id.equals(P_ID_REFERENCETYPE)) {
			
			// NOTE: this code will only be reached if user actually changes choice in reference type
			//	dropdown list, not if focus is merely gained and lost without changing the value
			
			// Grab reference's title cache
			String oldTitleCache = reference.getTitleCache();
			
			// Get reference's class
			int menuIndex;
			try {
				menuIndex = ((Integer) value).intValue();
			} catch (ClassCastException e) {
				// Leave method / no change on class cast error
				logger.error(e.getMessage());
				return;
			}			
			Class<?> referenceClass = (Class<?>)(getReferenceClassSet().toArray())[menuIndex];
			
			// Make a new reference using the class
			reference = makeNewReference(referenceClass);
			
			// Set new reference's title cache
			reference.setTitleCache(oldTitleCache);
		}
		
		// Author Team
		if (id.equals(P_ID_AUTHORTEAM)) {
			TeamOrPersonBase<?> authorTeam = reference.getAuthorTeam();
			if (authorTeam == null) {
				authorTeam = Team.NewInstance();
				reference.setAuthorTeam(authorTeam);
			}
			try {
				authorTeam.setTitleCache((String) value);
			} catch (ClassCastException e) {
				// Leave method / no change on class cast error
				logger.error(e.getMessage());
				return;
			}
		}
		
		// Citation
		if (id.equals(P_ID_URI)) {
			String uri = (String) value;
			reference.setUri(uri);
		}
		
		// Title: used by StrictReferenceBase
		if (id.equals(P_ID_TITLE)) {
			String title = null;
			try {
				title = (String) value;
			} catch (ClassCastException e) {
				// Leave method / no change on class cast error
				logger.error(e.getMessage());
				return;
			}
			invokeMethod(reference, "setTitle", title);
		}

		// Pages
		if (id.equals(P_ID_PAGES)) {
			
			invokeMethod(reference, "setPages", (String) value);
		}

		// Series
		if (id.equals(P_ID_SERIES)) {
			invokeMethod(reference, "setSeries", (String) value);
		}
		
		// Volume
		if (id.equals(P_ID_VOLUME)) {
			invokeMethod(reference, "setVolume", (String) value);
		}
		
		// Editor
		if (id.equals(P_ID_EDITOR)) {
			invokeMethod(reference, "setEditor", (String) value);
		}

		// Edition
		if (id.equals(P_ID_EDITION)) {
			invokeMethod(reference, "setEdition", (String) value);
		}
		
		// Place published
		if (id.equals(P_ID_PLACEPUBLISHED)) {
			invokeMethod(reference, "setPlacePublished", (String) value);
		}

		// Publisher
		if (id.equals(P_ID_PUBLISHER)) {
			invokeMethod(reference, "setPublisher", (String) value);
		}

		// ISSN
		if (id.equals(P_ID_ISSN)) {
			invokeMethod(reference, "setIssn", (String) value);
		}

		// Institution
		if (id.equals(P_ID_INSTITUTION)) {
			Institution institution = ((Report) reference).getInstitution();
			if (institution == null) {
				institution = Institution.NewInstance();
				((Report) reference).setInstitution(institution);
			}
			institution.setName((String) value);			
		}

		// Institution
		if (id.equals(P_ID_SCHOOL)) {
			Institution school = ((Thesis) reference).getSchool();
			if (school == null) {
				school = Institution.NewInstance();
				((Thesis) reference).setSchool(school);
			}
			school.setName((String) value);			
		}
		
		// ISBN
		if (id.equals(P_ID_ISBN)) {
			invokeMethod(reference, "setIsbn", (String) value);
		}

		// Organization
		if (id.equals(P_ID_ORGANIZATION)) {
			invokeMethod(reference, "setOrganization", (String) value);
		}
		
		// Date published
		if (id.equals(P_ID_DATEPUBLISHED)) {
						
			if (value instanceof String) {
				String year = (String) value;
				TimePeriod yearsPublished = null;
				
				// If not empty, format is either "xxxx" or "xxxx-xxxx", thanks to validator
				if (!year.equals("")) {
					
					// Convert to a TimePeriod
					yearsPublished = TimeHelper.convertTimePeriod(year);
					
					// Update reference's dataPublished field
					TimePeriod datePublished = reference.getDatePublished();
					
					if (datePublished == null) {
						reference.setDatePublished(yearsPublished);
					} else {
						
						// Any months or days should be left as is - only update years 
						datePublished.setStartYear(yearsPublished.getStartYear());
						datePublished.setEndYear(yearsPublished.getEndYear());
					}
				} else {
					reference.setDatePublished(null);
				}
			}
		}
		
		if (id.equals(P_ID_ABSTRACT)) {
			reference.setReferenceAbstract((String) value);
		}

		// Editable cache
		if (id.equals(P_ID_EDITABLECACHE)) {
			reference.setTitleCache((String) value);
		} else {
			reference.setTitleCache(reference.generateTitle(), reference.isProtectedTitleCache());
		}

		if (id.equals(P_ID_SERIESPART)) {
			((PrintedUnitBase) reference).setSeriesPart((String) value);
		}
		
		if (id.equals(P_ID_INREFERENCE)) {
			setInReference(value);
		}
		propertyChangeSupport.firePropertyChange(Resources.PROPERTY_SHEET_CHANGE, null, reference);
	}
	
	/**
	 * Takes a <code>referenceClass</code>, then executes either its <code>NewInstance()</code> 
	 * method or its constructor. Puts the resulting object in <code>this.reference</code> and
	 * returns it.
	 * <p>
	 * If unsuccessful, <code>this.reference</code> keeps its old value and is returned.
	 * @param referenceClass
	 * @return
	 */
	private ReferenceBase<?> makeNewReference(Class<?> referenceClass) {
				
		// Check whether the class has a method "NewInstance()";
		try {
			Method method = referenceClass.getMethod("NewInstance", null);
			reference = (ReferenceBase<?>) method.invoke(null, null);
		} catch (NoSuchMethodException e) {
			try {
				// If not, use its constructor
				Constructor<?> constructor = referenceClass.getConstructor(null);
				reference = (ReferenceBase<?>) constructor.newInstance(null);
			} catch (SecurityException e1) {
				e1.printStackTrace();
			} catch (NoSuchMethodException e1) {
				e1.printStackTrace();
			} catch (IllegalArgumentException e1) {
				e1.printStackTrace();
			} catch (InstantiationException e1) {
				e1.printStackTrace();
			} catch (IllegalAccessException e1) {
				e1.printStackTrace();
			} catch (InvocationTargetException e1) {
				e1.printStackTrace();
			}
		} catch (SecurityException e) {
			e.printStackTrace();
		} catch (IllegalArgumentException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		} catch (InvocationTargetException e) {
			e.printStackTrace();
		}
		
		return reference;
	}

	public String toString() {
		return CdmUtils.Nz(reference.getTitleCache());
	}

	private PropertyChangeSupport propertyChangeSupport = new PropertyChangeSupport(this);
	
	public void addPropertyChangeListener(
			PropertyChangeListener listener) {
		propertyChangeSupport.addPropertyChangeListener(listener);
	}
	
	public ReferenceBase<?> getReference() {
		return reference;
	}
}